// ===== uDMXCapture.cpp =====

/*
  Copyright (C) 2009 Lutz (ilLU[TZ]minator)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  Version 2 as published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details. The license is
  in the file "COPYING".

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

//#include <libintl.h>
#include "uDMXCapture.h"

// Headers for fltk
#include <FL/Fl_Preferences.H>
#include <FL/Fl_Help_View.H>
#include <FL/Fl_Color_Chooser.H>
#include <FL/Fl_Hold_Browser.H>
#include <FL/fl_ask.H>

// common headers
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>

#ifndef WIN32
// === Linux ...

#include <pthread.h>

#include "uDMX.xpm"
#include <FL/x.H>
#include <X11/xpm.h>
#else
// === WIN32
#include "uDMXCapturerc.h"
#include <process.h>
#endif

#include "uDMX_fkt.h"


// === idle() - update the dialog if nothing else to do
void idle(void * pUIvoid)
{
	pUIParams pUI = (pUIParams) pUIvoid ; 
	uint8_t *data = &pUI->dmx[0][0];
  int i ; 
  StringList::iterator t ;
  
  if (pUI->lbArtnet->value() != pUI->AnRunning)
    pUI->lbArtnet->value(pUI->AnRunning) ;  
  
  if (pUI->lbUDMX->value() != pUI->uDMXConnected)
    pUI->lbUDMX->value(pUI->uDMXConnected) ;
    

  // show actual dmx values
	for (i = 0; i < 20; i++)
	{
    pUI->progDMX[i]->value((data[i] * 100) / 255) ;
	}

  // show trace
  if (pUI->trcList.size() > 0)
  {
    // lock while changing list (threads change them too !)
		pUI->trcMutex.lock() ;
    
    for (t = pUI->trcList.begin(); t != pUI->trcList.end(); ++t)
    {
      string * strTrc = *t ;

      while (pUI->browDebug->size() > 100)
        pUI->browDebug->remove(1) ;
      
      pUI->browDebug->insert(pUI->browDebug->size() + 1, strTrc->c_str()) ; 
      pUI->browDebug->bottomline(pUI->browDebug->size()) ;
      
  		pUI->browDebug->redraw() ;

      delete strTrc ;
    }
    pUI->trcList.clear() ;
    // unlock after list changed
		pUI->trcMutex.unlock() ;
  }

  pUI->winMain->flush() ;

  SLEEP(30) ; // SAVE CPU time (dont update too often) !
  Fl::remove_idle(idle) ;  
  
  if (pUI->AnRestart)
    ArtNetRestart(pUI) ;
    
  return ;
}

char * DevStrTransform(char * cDev, bool bDecode = false)
{
  char * cDevOrg = cDev ;

  while(*cDev)
  {
    if (!bDecode && (*cDev == '\\'))
      *cDev = '[' ;
    if (bDecode && (*cDev == '['))
      *cDev = '\\' ;
  
    cDev++ ;
  }
  
  return cDevOrg ;
}


// === on LINUX pcap_next_ex() blocks 
// although using pcap_setnonblock() !
// Using pcap_dispatch() after select()
// works without blocking
void pcapCallBack(u_char *Param, 
                  const struct pcap_pkthdr *pktHdr, 
                  const u_char *pkt_data)
{
  int         Offset ;
  ArtNetDmx   *pArtnet ;

  UIParams *pPar = (UIParams *) Param ;

            // EthernetII + IP + UDP
  Offset = 14 + 20 + 8 ;

  // pointer to Art-Net data
  pArtnet = (ArtNetDmx *)(pkt_data + Offset) ;
  if ((strcmp((char *)pArtnet->id, "Art-Net") == 0) &&
      (pArtnet->universe == pPar->iDMXUniv) &&
      (pArtnet->opCode == ARTNET_DMX))
  {
    // Artnet-Data found !
    pPar->ulPacketCnt++ ;

    pPar->uiPacketLen = pArtnet->length_H * 0x0100 + pArtnet->length ;
    // store new data
    memcpy(pPar->dmx, pArtnet->data, pPar->uiPacketLen) ;
  }

  pPar->DmxLen  = pPar->uiPacketLen ;

  return ;
}

// Thread for receiving Art-Net data
#ifndef WIN32
void *pcapThread(void *Param)
#else
void pcapThread(void *Param)
#endif
{
  int                 res ;
  char                pcapErr[PCAP_ERRBUF_SIZE + 1] ;
  pcap_t              *hPcap ;
  struct bpf_program  filterCode ;
  bpf_u_int32         NetMask ;
  bpf_u_int32         IpAddr ;

  UIParams *pPar = (UIParams *) Param ;

  pPar->pcapError       = 0 ;
  // just started ... dont quit
  pPar->AnRunning       = true ;

  // open the network device
  if ((hPcap = pcap_open_live(pPar->cNetDev,  // name of the device
                              65536,          // portion of the packet to capture. 
                                              // 65536 grants that the whole packet will be
                                              //  captured on all the MACs.
                              1,              // promiscuous mode (nonzero means promiscuous)
                              5,              // timeout for running Art-Net
                              pcapErr         // error buffer
                              )) == NULL)
  {
    pPar->cUI->Trace(pPar, 0, "pcap_open_live(%s) Error %s",
                     pPar->cNetDev, pcapErr) ;
    pPar->pcapError   = true ;
    pPar->hPcapThread = 0 ;
#ifdef WIN32
    return ;
#else
    return NULL;
#endif
  }

  pPar->hPcap = hPcap ;

  if (pcap_setnonblock(hPcap, 1, pcapErr) == -1)
  {
    pPar->cUI->Trace(pPar, 1, "pcap_set_nonblock(%s) Error %s",
                     pPar->cNetDev, pcapErr) ;
  }

  /* ask pcap for the network address and mask of the device */
  if(pcap_lookupnet(pPar->cNetDev, &IpAddr, &NetMask, pcapErr) == -1)
  {
    pPar->cUI->Trace(pPar, 1, "pcap_lookupnet(%s) Error %s", 
                     pPar->cNetDev, pcapErr) ;
    pPar->pcapError   = true ;
    pPar->hPcapThread = 0 ;
#ifdef WIN32
    return ;
#else
    return NULL;
#endif
  }

  // Create filter for Art-Net
  if(pcap_compile(hPcap, &filterCode, "dst port 6454", 1, NetMask) < 0)
  {
    pPar->cUI->Trace(pPar, 1, "pcap_compile() Error %s", pcapErr) ;  
    pcap_close(hPcap) ;
    pPar->pcapError   = true ;
    pPar->hPcapThread = 0 ;
#ifdef WIN32
    return ;
#else
    return NULL;
#endif
  }

  //set the filter
  if(pcap_setfilter(hPcap, &filterCode) < 0)
  {
    pPar->cUI->Trace(pPar, 1, "pcap_setfilter()  Error") ;   
    pcap_close(hPcap) ;
    pPar->pcapError   = true ;
    pPar->hPcapThread = 0 ;
#ifdef WIN32
    return ;
#else
    return NULL;
#endif
  }

#ifndef WIN32
  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL) ;
  pthread_testcancel() ;
#endif

  // NO packets read so far ...
  pPar->ulPacketCnt = 0 ;


  //start the capture
#ifndef WIN32
  fd_set            readfds ;
  struct timeval    tv ;

  while(pPar->AnRunning)
  {
    FD_ZERO(&readfds);
    FD_SET(pcap_fileno(hPcap), &readfds) ;

    tv.tv_sec  = 0 ;
    tv.tv_usec = 5000 ;   /* 1000000 = 1 second */

    res = select(FD_SETSIZE, &readfds, NULL, NULL, &tv);

    switch(res)
    {
    // ERROR
    case(-1):
      pPar->cUI->Trace(pPar, 0, "%s / select error %d\n", __FUNCTION__, errno);
      break ;

    // timed out
    case(0):
      //dbgprintf(pPar, "%s / select: timed out\n", __FUNCTION__);
      break ;

    // data captured
    default:
      pcap_dispatch(hPcap, 1, pcapCallBack, (u_char *) pPar) ;
    }
  }
#else
  int                 Offset ;
  struct pcap_pkthdr  *header ;
  const  u_char       *pkt_data ;
  ArtNetDmx           *pArtnet ;

  while((res = pcap_next_ex( hPcap, &header, &pkt_data)) >= 0)
  {
    if (!pPar->AnRunning)
      // end the thread
      break ;

    if (!res)
    {
      Sleep(1) ;
      continue ;
    }

          // EthernetII + IP + UDP
    Offset = 14 + 20 + 8 ;

    // pointer to Art-Net data
    pArtnet = (ArtNetDmx *)(pkt_data + Offset) ;
    if ((strcmp((char *)pArtnet->id, "Art-Net") == 0) &&
        (pArtnet->universe == pPar->iDMXUniv) &&
        (pArtnet->opCode == ARTNET_DMX))
    {
      // Artnet-Data found !
      pPar->ulPacketCnt++ ;

      pPar->uiPacketLen = pArtnet->length_H * 0x0100 + pArtnet->length ;
      
      // store new data
      pPar->DmxLen  = pPar->uiPacketLen ;
      memcpy(pPar->dmx, pArtnet->data, pPar->uiPacketLen) ;
      
      pPar->cUI->Trace(pPar, 9, "rcvd %d bytes. %02X %02X %02X %02X %02X %02X %02X %02X ", 
                      pPar->DmxLen, pArtnet->data[0], pArtnet->data[1], 
                      pArtnet->data[2], pArtnet->data[3], 
                      pArtnet->data[4], pArtnet->data[5], 
                      pArtnet->data[6], pArtnet->data[7]) ;          
    }
  }
#endif

  // end capturing
  // actually we DO NOT arrive here because the thread is cancelled !
  pcap_freecode(&filterCode) ;
  pcap_close(hPcap) ;
  pPar->cUI->Trace(pPar, 8, "%s ended for >%s<\n", __FUNCTION__, pPar->cNetDev) ;

  pPar->hPcapThread = 0 ;
#ifdef WIN32
    return ;
#else
    return NULL;
#endif
}


// === Try to find at least ONE network device with Art-Net packets
int ArtNetDeviceDetect(UIParams *pPar)
{
  bool        bFound = false ;
  int         i = 0 ;
  char        pcapErr[PCAP_ERRBUF_SIZE + 1] ;
  pcap_if_t   *allPcapDevs ;
  pcap_if_t   *pcapDev ;
#ifndef WIN32
  int         res ;
#endif

  /* Retrieve the device list */
  if(pcap_findalldevs(&allPcapDevs, pcapErr) == -1)
  {
    pPar->cUI->Trace(pPar, 1, "pcap_findalldevs()  Error %s", pcapErr) ;  
    return false ;
  }

  /* Scan the list printing every entry */
  for(pcapDev = allPcapDevs; pcapDev != NULL; pcapDev = pcapDev->next)
  {
    i++ ;
    pPar->cUI->Trace(pPar, 8, "%d. %s", i, pcapDev->name) ;
    if (pcapDev->description)
      pPar->cUI->Trace(pPar, 8, "   %s", pcapDev->description);
  }
  
  if (i == 0)
  {
    pPar->cUI->Trace(pPar, 0, "No interfaces found! Make sure Pcap is installed") ;    
    return false ;
  }

  // Found some inferfaces -> look for Art-Net Packets
  for(pcapDev = allPcapDevs; pcapDev != NULL; pcapDev = pcapDev->next)
  {
    //
    pPar->ulPacketCnt = 0 ;
    pPar->cUI->Trace(pPar, 6, "checking for Artnet-packets on >%s< ...", pcapDev->name) ;   
    strcpy(pPar->cNetDev, pcapDev->name) ;   

    // create the thread to get packets ...
#ifndef WIN32
    if ((res = pthread_create(&pPar->hPcapThread, NULL, pcapThread, pPar)) != 0)
    {
      pPar->cUI->Trace(pPar, 0, "pthread_create failed. Error %d", res) ;
    }
#else
    pPar->hPcapThread = (HANDLE) _beginthread(pcapThread, 0, (void *)pPar) ;
#endif

    // wait 1 second for packets to arrive
    // typically there should be several within 1 second
    idle(pPar) ;    
    SLEEP(pPar->iPCTo) ;
    pPar->AnRunning = false ;
    SLEEP(100) ;

    if (pPar->hPcapThread)
    {
#ifndef WIN32
      // cancel thread
      //pcap_breakloop(pPar->hPcap) ;
      if ((res = pthread_cancel(pPar->hPcapThread)) != 0)
        pPar->cUI->Trace(pPar, 0,  "pthread_cancel failed. Error %d", res) ;
#else
      TerminateThread(pPar->hPcapThread, 0) ;
#endif
    }

    // has the thread received something ?
    if (pPar->ulPacketCnt)
    {
      bFound = true ;
      pPar->cUI->Trace(pPar, 6, "%d Art-Net packets received on %s", 
                      pPar->ulPacketCnt, pPar->cNetDev) ;  

      // select the stored value
      pPar->ddPCDev->clear() ;
      strcpy(pPar->cNetDev2, pPar->cNetDev) ;
      pPar->ddPCDev->add(DevStrTransform(pPar->cNetDev2)) ;  
      pPar->ddPCDev->value(0) ;      
                      
      break ;
    }
  } // for(pcapDev = allPcapDevs; pcapDev != NULL; pcapDev = pcapDev->next)

  /* Free the device list */
  pcap_freealldevs(allPcapDevs) ;

  return (bFound) ;
}


int ArtNetEnd(UIParams *pPar)
{
#ifndef WIN32
  int   res ;
  void  *thStatus ;
#endif

  pPar->AnRunning = false ;

#ifndef WIN32
  if (pPar->hPcapThread)
  {
    // cancel the thread
    if ((res = pthread_cancel(pPar->hPcapThread)) != 0)
      pPar->cUI->Trace(pPar, 6, "pthread_cancel failed. Error %d\n", res) ;

    if ((res = pthread_join(pPar->hPcapThread, &thStatus)) != 0)
      pPar->cUI->Trace(pPar, 6, "pthread_join failed. Error %d\n", res) ;

    if (thStatus != PTHREAD_CANCELED)
      pPar->cUI->Trace(pPar, 6, "Unexpected thread status\n") ;
  }
#else
  if(pPar->hPcapThread)
    TerminateThread(pPar->hPcapThread, 0) ;
#endif

  return true ;
}

// === do everything that is necessary to output data:
// o open USB device (uDMX)
// o look for network device with Art-Net data
int ArtNetRestart(UIParams *pPar)
{
#ifndef WIN32
  int res ;
#endif

  pPar->AnRestart = false ;

  // clear Info for interface
  pPar->cNetDev[0] = 0 ;

  ArtNetEnd(pPar) ;

  if (!strlen(pPar->cNetDev))
  {
    // No device given -> try to detect
    pPar->cUI->Trace(pPar, 6, "trying to detect device with Art-Net packets %s", pPar->cNetDev) ; 
    if (!ArtNetDeviceDetect(pPar))
    {
      strcpy(pPar->cNetDev, "") ;
      pPar->cUI->Trace(pPar, 6, "NO Art-Net packets for Universe >%d< found",
                                pPar->iDMXUniv) ; 
    }
    else
      pPar->AnRunning = true ;
  }

  if (strlen(pPar->cNetDev))
  {
    // create the thread to get packets ...
#ifndef WIN32
    if ((res = pthread_create(&pPar->hPcapThread, NULL, pcapThread, pPar)) != 0)
    {
      pPar->cUI->Trace(pPar, 0, "pthread_create failed. Error %d", res) ;
    }
#else
    pPar->hPcapThread = (HANDLE) _beginthread(pcapThread, 0, (void *)pPar) ;
#endif
    pPar->cUI->Trace(pPar, 6, "Reading Art-Net from >%s<", pPar->cNetDev) ;
  }

  return true ;
}




// ColorSelectBG() - select background color
void ColorSelectBG(Fl_Widget *, void * pUIvoid)
{
//  UIParams *pUI = (UIParams *)pUIvoid ;  
  Fl_Color c = FL_BACKGROUND_COLOR ;  
  uchar r, g, b; 
  
  Fl::get_color(c, r, g, b) ;
  
  if (fl_color_chooser(0, r, g, b))
  {
    Fl::set_color(c, r, g, b) ; 
    Fl::redraw();
  }

  return ;
}
// ColorSelectFG() - select foreground color
void ColorSelectFG(Fl_Widget *, void * pUIvoid)
{
//  UIParams *pUI = (UIParams *)pUIvoid ;  
  Fl_Color c = FL_FOREGROUND_COLOR ;  
  uchar r, g, b; 
  
  Fl::get_color(c, r, g, b) ;
  
  if (fl_color_chooser(0, r, g, b))
  {
    Fl::set_color(c, r, g, b) ; 
    Fl::redraw();
  }

  return ;
}
// ColorSelectIN() - select inputs background color
void ColorSelectIN(Fl_Widget *, void * pUIvoid)
{
//  UIParams *pUI = (UIParams *)pUIvoid ;  
  Fl_Color c = FL_BACKGROUND2_COLOR ;  
  uchar r, g, b; 
  
  Fl::get_color(c, r, g, b) ;
  
  if (fl_color_chooser(0, r, g, b))
  {
    Fl::set_color(c, r, g, b) ; 
    Fl::redraw();
  }

  return ;
}
// ColorSelectPG() - select progress bars color
void ColorSelectPG(Fl_Widget *, void * pUIvoid)
{
//  UIParams *pUI = (UIParams *)pUIvoid ;  
  Fl_Color c = (Fl_Color)4 ;  
  uchar r, g, b; 
  
  Fl::get_color(c, r, g, b) ;
  
  if (fl_color_chooser(0, r, g, b))
  {
    Fl::set_color(c, r, g, b) ; 
    Fl::redraw();
  }

  return ;
}

void VerboseChange(Fl_Widget *, void * pUIvoid)
{
	pUIParams pUI = (pUIParams) pUIvoid ;
  pUI->iVerbose = (int)pUI->spinVerbose->value() ;
}
void DMXUnivChange(Fl_Widget *, void * pUIvoid)
{
	pUIParams pUI = (pUIParams) pUIvoid ;
  pUI->iDMXUniv = (int)pUI->spinDMXUniv->value() ;
  
  // transfer parameter to node
  //pUI->port_addr  = pUI->iDMXUniv ;	  
  // restart artnet
  //###artnetStop(pUI) ;
  pUI->AnRestart = true ;    
}
void TimeBreakChange(Fl_Widget *, void * pUIvoid)
{
	pUIParams pUI = (pUIParams) pUIvoid ;
  pUI->iUDMXTb  = (int)pUI->spinUDMXTb->value() ;
}
void TimeMarkChange(Fl_Widget *, void * pUIvoid)
{
	pUIParams pUI = (pUIParams) pUIvoid ;
  pUI->iUDMXTm  = (int)pUI->spinUDMXTm->value() ;
}
void TimeGapChange(Fl_Widget *, void * pUIvoid)
{
	pUIParams pUI = (pUIParams) pUIvoid ;
  pUI->iUDMXTg  = (int)pUI->spinUDMXTg->value() ;
}

void Debug2Clip(Fl_Widget *, void * pUIvoid)
{
	pUIParams pUI = (pUIParams) pUIvoid ;
  const char * cIn ;
  char *cOut = NULL ;
  int sizeOld = 0, sizeNew ;

  for (int i = 1; i <= pUI->browDebug->size(); i++)
  {
    cIn   = pUI->browDebug->text(i) ;
    if (!cIn)
      continue ;

    sizeNew = sizeOld + strlen(cIn) + 3 ;
		cOut  = (char *)realloc(cOut, sizeNew) ;

		cOut[sizeOld]	= 0 ;

    strcat(cOut, "\r\n") ;
    strcat(cOut, cIn) ;

		sizeOld = sizeNew ;
  }

  Fl::copy(cOut, strlen(cOut) + 1, 1) ;
 
  free(cOut) ;
}

// prefsFile() - create filename for prefs depending on parameter cSerial
bool PrefsFile(pUIParams pUI, char *cFile)
{
	strcpy(cFile, "uDMXCapture") ;
  if (pUI->bSerial || 
      strlen(pUI->cConfig))
  {
    strcat(cFile, "_") ;
    if (!strlen(pUI->cConfig))
      strcat(cFile, pUI->cSerial) ;
    else
      strcat(cFile, pUI->cConfig) ;    
  }

	return true ;
}

// LayoutSave() - save data responsible for dialogs layout
void LayoutSave(pUIParams pUI)
{
  // Path for Settings
	char cFile[MAX_PATH] ;

  PrefsFile(pUI, (char *)&cFile) ;
  Fl_Preferences prefs( Fl_Preferences::USER, "ilLUTZminator.de", cFile);
  
  // Position
  Fl_Preferences prefPos( prefs, "Layout" ) ; 
  prefPos.set("Left", pUI->winMain->x()) ; 
  prefPos.set("Top",  pUI->winMain->y()) ;   
  
  prefPos.set("Scheme", pUI->ddScheme->value()) ;
  
  prefPos.set("ColorBG", (int)Fl::get_color(FL_BACKGROUND_COLOR)) ;  
  prefPos.set("ColorFG", (int)Fl::get_color(FL_FOREGROUND_COLOR)) ;    
  prefPos.set("ColorIN", (int)Fl::get_color(FL_BACKGROUND2_COLOR)) ;      
  prefPos.set("ColorPG", (int)Fl::get_color((Fl_Color)4)) ;        
  
  return ;
}

// SchemeSet() - set scheme on dropdowns (Fl_Choice) selection
void SchemeSet(Fl_Choice * oChoice, void * pUIvoid)
{
  Fl::scheme(oChoice->text(oChoice->value()));
  return ;
}

// ShutdownQuery() - dont stop program without confirmation
void ShutdownQuery(Fl_Widget *, void * pUIvoid)
{
  UIParams *pUI = (UIParams *)pUIvoid ;

  int rep = fl_choice("uDMX capture: Do you want to quit ?", 
                      "Yes", "No", 0);
  if (rep == 0) 
  {
    LayoutSave(pUI) ;
    
    pUI->bShutdown = true ;
    pUI->AnRunning = false ;
		//###artnetStop(pUI) ;
				
		pUI->cUI->Trace(pUI, 6, "waiting for threads to end") ;	
    SLEEP(2000) ;
    
    exit(0);
  }

  return ;
}

// Value2SingleBit() - changes value to value that is limited to
// a value that has only one bit set (1, 2, 4, 8, 16 ... 512)
int Value2SingleBit(int iValue, pUIParams pUI)
{
  int iBit = 0 ;
  
  while(iValue > 0)
  {
    iValue = iValue >> 1 ;
    iBit ++ ;
//		pUI->cUI->Trace(pUI, 0, "iValue: %d iBit %d", iValue, iBit) ;	    
  }
  iValue = (iBit ? (1 << (iBit - 1)) : 0) ;

  return iValue ;
}

// SettingsSave() - save all settings needed for program (except layout)
void SettingsSave(Fl_Widget *, void * pUIvoid)
{
  UIParams *pUI = (UIParams *)pUIvoid ;

  // Path for Settings
	char cFile[MAX_PATH] ;

  PrefsFile(pUI, (char *)&cFile) ;
  Fl_Preferences prefs( Fl_Preferences::USER, "ilLUTZminator.de", cFile);

  // Common settings
  Fl_Preferences prefCommon( prefs, "Common" ) ;
  pUI->iVerbose = (int)pUI->spinVerbose->value() ;
  prefCommon.set( "Verbose", pUI->iVerbose);
  
  // Artnet settings
  Fl_Preferences prefArtnet( prefs, "Capture" ) ;
  strcpy(pUI->cNetDev, pUI->ddPCDev->input()->value() ) ;
  prefArtnet.set( "Device", pUI->cNetDev ); 
  pUI->iPCTo  = (int)pUI->spinPCTo->value() ;
  prefArtnet.set( "ScanTO", pUI->iPCTo );
  
  // DMX
  Fl_Preferences prefDMX( prefs, "DMX" ) ;
  pUI->iDMXUniv   = (int)pUI->spinDMXUniv->value() ;
  prefDMX.set( "Universe", pUI->iDMXUniv );
  
  pUI->iDMXChans  = (int)pUI->spinDMXChans->value() ;
  pUI->iDMXChans  = Value2SingleBit(pUI->iDMXChans, pUI) ;
  pUI->spinDMXChans->value(pUI->iDMXChans) ;
  prefDMX.set( "Channels", pUI->iDMXChans );  

  // uDMX
  Fl_Preferences prefUDMX( prefs, "uDMX" ) ;
  
  pUI->iUDMXBs    = (int)pUI->spinUDMXBs->value() ; 
  pUI->iUDMXBs    = Value2SingleBit(pUI->iUDMXBs, pUI) ;
  pUI->spinUDMXBs->value(pUI->iUDMXBs) ;
  prefUDMX.set( "BlockSize", pUI->iUDMXBs );
  
  pUI->iUDMXTb    = (int)pUI->spinUDMXTb->value() ;
  prefUDMX.set( "TimeBreak", pUI->iUDMXTb );
  pUI->iUDMXTm    = (int)pUI->spinUDMXTm->value() ;
  prefUDMX.set( "TimeMark", pUI->iUDMXTm );
  pUI->iUDMXTg    = (int)pUI->spinUDMXTg->value() ;
  prefUDMX.set( "TimeIBGap", pUI->iUDMXTg );  
  
  // calculated values
  //pUI->ip_addr	= (strlen(pUI->cANIp) != 0) ? pUI->cANIp : NULL ;
  //pUI->bind2Ip	= (strlen(pUI->cANIp) != 0) ;   

  // transfer parameter to node
  //pUI->port_addr  = pUI->iDMXUniv ;	  

  // restart artnet
  //###artnetStop(pUI) ;
  pUI->AnRestart = true ;  
  //ArtNetRestart(pUI) ;
  
  return ;
}

// ButtonUnuse() - make button unusable
// if button is pressed it does NOT change state
static void ButtonUnuse(Fl_Widget * pBtn, void *)
{
  Fl_Light_Button *pL = (Fl_Light_Button *)pBtn ;
  pL->value(!pL->value()) ;
  return ;
}

// Trace() - output in debug-window
bool UserInterface::Trace(pUIParams pUI, int level, const char *fmt, ...)
{
  char Buffer[1024] ;

  if ((level <= pUI->iVerbose) && (fmt))
  {
    va_list arglist ;
    va_start(arglist, fmt) ;	

#ifdef FL_LOCK
  	Fl::lock() ;
#endif

		snprintf(Buffer, sizeof(Buffer), "%04d ", pUI->TrcID) ;
		pUI->TrcID++ ;
		pUI->TrcID = pUI->TrcID % 10000 ;
		
	  vsnprintf(Buffer + 5, sizeof(Buffer) - 5, fmt, arglist);

    /* NO CHANGES OF FLTK ELEMENT FROM THREAD
       THIS LEADS TO CHRASHES WITH X11
    while (pUI->browDebug->size() > 100)
      pUI->browDebug->remove(1) ;
      
    pUI->browDebug->insert(pUI->browDebug->size() + 1, Buffer) ; 
    pUI->browDebug->bottomline(pUI->browDebug->size()) ;
      
		pUI->browDebug->redraw() ;
    */
    // STORE STRING IN LIST AND SHOW LIST IN IDLE FUNCTION
    string *str = new string(Buffer) ;

    // lock while changing list 
    // (main process and threads do tracing / change list ...)
		pUI->trcMutex.lock() ;
    pUI->trcList.push_back(str) ;
    // unlock after changing list
		pUI->trcMutex.unlock() ;

#ifdef FL_LOCK
	  Fl::awake(idle, pUI) ;
  	Fl::unlock() ;
#endif
    
	  va_end(arglist) ;
  }

  return true ;
}

// SettingsRead() - get all parameters for program (except layout)
bool UserInterface::SettingsRead(pUIParams pUI)
{
  // Path for Settings
	char cFile[MAX_PATH] ;

  PrefsFile(pUI, (char *)&cFile) ;
  Fl_Preferences prefs( Fl_Preferences::USER, "ilLUTZminator.de", cFile);

  //char path[ FL_PATH_MAX ];
  //refs.getUserdataPath( path, sizeof(path) );
  
  // Common settings
  Fl_Preferences prefCommon( prefs, "Common" ) ;
  prefCommon.get( "Verbose", pUI->iVerbose, 0) ;
  pUI->spinVerbose->value(pUI->iVerbose) ;  
  
  // Artnet settings
  Fl_Preferences prefArtnet( prefs, "Capture" ) ; 
  
  // Device
  prefArtnet.get( "Device", pUI->cNetDev, "", MAX_PATH - 1);
  if (strlen(pUI->cNetDev) > 0)
  {
    // select the stored value
    strcpy(pUI->cNetDev2, pUI->cNetDev) ;
    pUI->ddPCDev->add(DevStrTransform(pUI->cNetDev2)) ;  
    pUI->ddPCDev->value(0) ;    
  }
  prefArtnet.get( "ScanTO", pUI->iPCTo, 4500) ;
  pUI->spinPCTo->value(pUI->iPCTo) ;  
  
  // DMX
  Fl_Preferences prefDMX( prefs, "DMX" ) ;
  prefDMX.get( "Universe", pUI->iDMXUniv, 0) ;
  pUI->spinDMXUniv->value(pUI->iDMXUniv) ;
  prefDMX.get( "Channels", pUI->iDMXChans, 512) ; 
  pUI->spinDMXChans->value(pUI->iDMXChans) ;   

  // uDMX
  Fl_Preferences prefUDMX( prefs, "uDMX" ) ;
  prefUDMX.get( "BlockSize", pUI->iUDMXBs, 128) ;
  pUI->spinUDMXBs->value(pUI->iUDMXBs) ;
  prefUDMX.get( "TimeBreak", pUI->iUDMXTb, 88) ;
  pUI->spinUDMXTb->value(pUI->iUDMXTb) ;
  prefUDMX.get( "TimeMark", pUI->iUDMXTm, 8) ;
  pUI->spinUDMXTm->value(pUI->iUDMXTm) ;
  prefUDMX.get( "TimeIBGap", pUI->iUDMXTg, 0) ;
  pUI->spinUDMXTg->value(pUI->iUDMXTg) ;   
  
  // calculated values
  //pUI->ip_addr	= (strlen(pUI->cANIp) != 0) ? pUI->cANIp : NULL ;
  //pUI->bind2Ip	= (strlen(pUI->cANIp) != 0) ; 

  // initialisation of structure
  //pUI->port_addr  = pUI->iDMXUniv ;
  //pUI->num_ports  = 1 ;
  pUI->DmxLen     = 0 ;

  memset(&pUI->dmx[0], 0, ARTNET_MAX_DMX +1) ;
  
  pUI->bShutdown  = false ;
  
  pUI->AnRunning  = true ;
  pUI->AnRestart  = true ;
  
  pUI->uDMXConnected = false ;
   
  pUI->cNetDev[0] = 0 ; 
  
	pUI->trcList.clear() ;
  pUI->TrcID      = 0 ;

  pUI->hPcapThread = NULL ;
  
  return true ;
}

// LayoutRead() - get parameters for layout
bool UserInterface::LayoutRead(pUIParams pUI)
{
  int X, Y, iScheme, iColor ;
  // Path for Settings
	char cFile[MAX_PATH] ;

  PrefsFile(pUI, (char *)&cFile) ;
  Fl_Preferences prefs( Fl_Preferences::USER, "ilLUTZminator.de", cFile);
  
  // Position
  Fl_Preferences prefPos( prefs, "Layout" ) ; 
  prefPos.get("Left", X, 100) ;
  prefPos.get("Top", Y, 100) ;  
  pUI->winMain->position(X, Y) ;

  // Scheme
  prefPos.get("Scheme", iScheme, 0) ;
  pUI->ddScheme->value(iScheme) ;
  SchemeSet(pUI->ddScheme, pUI) ;
  
  // Color
  prefPos.get("ColorBG", iColor, (int)Fl::get_color(FL_BACKGROUND_COLOR)) ; 
  Fl::set_color(FL_BACKGROUND_COLOR, (unsigned) iColor) ; 
  prefPos.get("ColorFG", iColor, (int)Fl::get_color(FL_FOREGROUND_COLOR)) ; 
  Fl::set_color(FL_FOREGROUND_COLOR, (unsigned) iColor) ; 
  prefPos.get("ColorIN", iColor, (int)Fl::get_color(FL_BACKGROUND2_COLOR)) ; 
  Fl::set_color(FL_BACKGROUND2_COLOR, (unsigned) iColor) ; 
  prefPos.get("ColorPG", iColor, (int)Fl::get_color((Fl_Color)4)) ; 
  Fl::set_color((Fl_Color)4, (unsigned) iColor) ; 
  
  Fl::redraw();  
  
  return true ;
}

// make_window() - create dialog
Fl_Double_Window* UserInterface::make_window(UIParams *pUI)
{
  Fl_Double_Window* w;
  { Fl_Double_Window* o = new Fl_Double_Window(424, 316, _("uDMX capture"));
    w = o;
    o->user_data((void*)(this));
    o->align(Fl_Align(FL_ALIGN_CLIP|FL_ALIGN_INSIDE));
    { Fl_Tabs* o = new Fl_Tabs(0, 0, 450, 316);
      { Fl_Group* o = new Fl_Group(0, 26, 425, 284, _("&uDMX"));
        { pUI->browDebug = new Fl_Hold_Browser(4, 120, 418, 190);
          pUI->browDebug->callback((Fl_Callback*)Debug2Clip, (void *)pUI);
          pUI->browDebug->tooltip(_("Mouseclick copies contents to clipboard"));
        } // Fl_Browser* o
        { pUI->lbArtnet = new Fl_Light_Button(10, 34, 65, 20, _("Artnet"));
          pUI->lbArtnet->selection_color((Fl_Color)2);
          pUI->lbArtnet->tooltip(_("artnet:\nstartet OK if light green\n(no guarantee to receive data)"));           
          // Dont use deactivate() as color is nearly not recognizable
          // pUI->lbArtnet->deactivate() ;
          // Make Button unusable instead          
          pUI->lbArtnet->callback((Fl_Callback*)ButtonUnuse);                     
        } // Fl_Light_Button* o
        { pUI->lbUDMX = new Fl_Light_Button(90, 34, 65, 20, _("uDMX"));
          pUI->lbUDMX->selection_color((Fl_Color)2);
          pUI->lbUDMX->tooltip(_("uDMX:\nstartet OK if light green"));           
          pUI->lbUDMX->callback((Fl_Callback*)ButtonUnuse);           
          //pUI->lbUDMX->value(1) ;          
        } // Fl_Light_Button* o
        { pUI->progDMX[0] = new Fl_Progress(10, 64, 30, 20, _("001"));
          pUI->progDMX[0]->selection_color((Fl_Color)4);
          pUI->progDMX[0]->tooltip(_("data of first DMX channel"));           
        } // Fl_Progress* o
        { pUI->progDMX[1] = new Fl_Progress(45, 64, 30, 20, _("002"));
          pUI->progDMX[1]->selection_color((Fl_Color)4);
        } // Fl_Progress* o
        { pUI->progDMX[2] = new Fl_Progress(80, 64, 30, 20, _("003"));
          pUI->progDMX[2]->selection_color((Fl_Color)4);
        } // Fl_Progress* o
        { pUI->progDMX[3] = new Fl_Progress(115, 64, 30, 20, _("004"));
          pUI->progDMX[3]->selection_color((Fl_Color)4);
        } // Fl_Progress* o
        { pUI->progDMX[4] = new Fl_Progress(150, 64, 30, 20, _("005"));
          pUI->progDMX[4]->selection_color((Fl_Color)4);
        } // Fl_Progress* o
        { pUI->progDMX[5] = new Fl_Progress(185, 64, 30, 20, _("006"));
          pUI->progDMX[5]->selection_color((Fl_Color)4);
        } // Fl_Progress* o
        { pUI->progDMX[6] = new Fl_Progress(220, 64, 30, 20, _("007"));
          pUI->progDMX[6]->selection_color((Fl_Color)4);
        } // Fl_Progress* o
        { pUI->progDMX[7] = new Fl_Progress(255, 64, 30, 20, _("008"));
          pUI->progDMX[7]->selection_color((Fl_Color)4);
        } // Fl_Progress* o
        { pUI->progDMX[8] = new Fl_Progress(290, 64, 30, 20, _("009"));
          pUI->progDMX[8]->selection_color((Fl_Color)4);
        } // Fl_Progress* o
        { pUI->progDMX[9] = new Fl_Progress(325, 64, 30, 20, _("010"));
          pUI->progDMX[9]->selection_color((Fl_Color)4);
        } // Fl_Progress* o
        { pUI->progDMX[10] = new Fl_Progress(10, 90, 30, 20, _("011"));
          pUI->progDMX[10]->selection_color((Fl_Color)4);
        } // Fl_Progress* o
        { pUI->progDMX[11] = new Fl_Progress(45, 90, 30, 20, _("012"));
          pUI->progDMX[11]->selection_color((Fl_Color)4);
        } // Fl_Progress* o
        { pUI->progDMX[12] = new Fl_Progress(80, 90, 30, 20, _("013"));
          pUI->progDMX[12]->selection_color((Fl_Color)4);
        } // Fl_Progress* o
        { pUI->progDMX[13] = new Fl_Progress(115, 90, 30, 20, _("014"));
          pUI->progDMX[13]->selection_color((Fl_Color)4);
        } // Fl_Progress* o
        { pUI->progDMX[14] = new Fl_Progress(150, 90, 30, 20, _("015"));
          pUI->progDMX[14]->selection_color((Fl_Color)4);
        } // Fl_Progress* o
        { pUI->progDMX[15] = new Fl_Progress(185, 90, 30, 20, _("016"));
          pUI->progDMX[15]->selection_color((Fl_Color)4);
        } // Fl_Progress* o
        { pUI->progDMX[16] = new Fl_Progress(220, 90, 30, 20, _("017"));
          pUI->progDMX[16]->selection_color((Fl_Color)4);
        } // Fl_Progress* o
        { pUI->progDMX[17] = new Fl_Progress(255, 90, 30, 20, _("018"));
          pUI->progDMX[17]->selection_color((Fl_Color)4);
        } // Fl_Progress* o
        { pUI->progDMX[18] = new Fl_Progress(290, 90, 30, 20, _("019"));
          pUI->progDMX[18]->selection_color((Fl_Color)4);
        } // Fl_Progress* o
        { pUI->progDMX[19] = new Fl_Progress(325, 90, 30, 20, _("020"));
          pUI->progDMX[19]->selection_color((Fl_Color)4);
        } // Fl_Progress* o
        o->end();
      } // Fl_Group* o
      { Fl_Group* o = new Fl_Group(0, 26, 425, 289, _("&Setup"));
        o->hide();
        { Fl_Group* o = new Fl_Group(40, 31, 378, 49, _("Common"));
          o->labelfont(1);
          o->align(Fl_Align(FL_ALIGN_TOP_LEFT|FL_ALIGN_INSIDE));
          { pUI->spinVerbose = new Fl_Spinner(120, 56, 45, 20, _("Verbose"));
            pUI->spinVerbose->minimum(0);
#ifdef FL_LOCK
            pUI->spinVerbose->maximum(0);
#else
            pUI->spinVerbose->maximum(9);
#endif
            pUI->spinVerbose->value(0);
            pUI->spinVerbose->tooltip(_("verbose:\nthe higher the value,\nthe more info you will see on uDMX page"));            
            pUI->spinVerbose->callback((Fl_Callback*)VerboseChange, (void *)pUI);                
          } // Fl_Spinner* o
          { pUI->ddScheme = new Fl_Choice(235, 55, 85, 20, _("Scheme"));
            pUI->ddScheme->down_box(FL_BORDER_BOX);
            pUI->ddScheme->add("none");
            pUI->ddScheme->add("gtk+");
            pUI->ddScheme->add("gleam");            
            pUI->ddScheme->add("plastic");
            pUI->ddScheme->callback((Fl_Callback *)SchemeSet, pUI);
            pUI->ddScheme->tooltip(_("scheme:\nchanges the programs look"));            
            
          } // Fl_Choice* o          
          { Fl_Button* o = new Fl_Button(330, 55, 20, 20, _("C"));
            o->tooltip(_("color:\nchanges the programs background color.\nbetter NOT choose same color like foreground ;-)"));           
            o->callback((Fl_Callback*)ColorSelectBG, (void *)pUI);            
          } // Fl_Button* o
          { Fl_Button* o = new Fl_Button(352, 55, 20, 20, _("C"));
            o->tooltip(_("color:\nchanges the programs foreground color.\nbetter NOT choose same color like background ;-)"));           
            o->callback((Fl_Callback*)ColorSelectFG, (void *)pUI); 
            o->color(FL_FOREGROUND_COLOR);
            o->labelcolor(FL_BACKGROUND_COLOR);
          } // Fl_Button* o
          { Fl_Button* o = new Fl_Button(374, 55, 20, 20, _("C"));
            o->tooltip(_("color:\nchanges the inputs background color.\nbetter NOT choose same color like foreground ;-)")); 
            o->callback((Fl_Callback*)ColorSelectIN, (void *)pUI);             
            o->color(FL_BACKGROUND2_COLOR);
          } // Fl_Button* o          
          { Fl_Button* o = new Fl_Button(396, 55, 20, 20, _("C"));
            o->tooltip(_("color:\nchanges the progress bar color.\n")); 
            o->callback((Fl_Callback*)ColorSelectPG, (void *)pUI);             
            o->color((Fl_Color)4);
          } // Fl_Button* o          
          o->end();
        } // Fl_Group* o        
        { Fl_Group* o = new Fl_Group(40, 86, 345, 60, _("Capture"));
          o->labelfont(1);
          o->align(Fl_Align(FL_ALIGN_TOP_LEFT|FL_ALIGN_INSIDE));
          { pUI->ddPCDev = new Fl_Input_Choice(95, 102, 320, 20, _("Device"));
            pUI->ddPCDev->tooltip(_("device:\nfirst network device with detected artnet-data"));  
          } // Fl_Input_Choice* o
          { pUI->spinPCTo = new Fl_Spinner(96, 124, 64, 20, _("ScanTime"));
            pUI->spinPCTo->minimum(500);
            pUI->spinPCTo->maximum(5000);
            pUI->spinPCTo->step(100);
            pUI->spinPCTo->tooltip(_("scantime:\n"
                                     "time to detect artnet-data\n"
                                     "secure > 4000\n"
                                     "program will freeze for this time\n"
                                     "multiplied with the number of network adapters\n"
                                     "when starting or saving !"));  
          } // Fl_Spinner* o
          o->end();
        } // Fl_Group* o
        { Fl_Group* o = new Fl_Group(40, 156, 345, 55, _("DMX"));
          o->labelfont(1);
          o->align(Fl_Align(FL_ALIGN_TOP_LEFT|FL_ALIGN_INSIDE));
          { pUI->spinDMXUniv = new Fl_Spinner(130, 172, 45, 20, _("Universe"));
            pUI->spinDMXUniv->minimum(0);
            pUI->spinDMXUniv->maximum(15);
            pUI->spinDMXUniv->value(0);
            pUI->spinDMXUniv->tooltip(_("universe:\nfirst universe to receive data.\ndefault 0"));  
            pUI->spinDMXUniv->callback((Fl_Callback*)DMXUnivChange, (void *)pUI);  
          } // Fl_Spinner* o
          { pUI->spinDMXChans = new Fl_Spinner(275, 172, 45, 20, _("Channels"));
            pUI->spinDMXChans->minimum(1);          
            pUI->spinDMXChans->maximum(512);
            pUI->spinDMXChans->value(512);
            pUI->spinDMXChans->tooltip(_("channels:\nno of channels to use.\ndefault 512\nvalid values 1,2,4,8,16...512"));              
          } // Fl_Spinner* o
          o->end();
        } // Fl_Group* o
        { Fl_Group* o = new Fl_Group(40, 213, 345, 67, _("uDMX"));
          o->labelfont(1);
          o->align(Fl_Align(FL_ALIGN_TOP_LEFT|FL_ALIGN_INSIDE));
          { pUI->spinUDMXBs = new Fl_Spinner(130, 235, 45, 20, _("USBBlock"));
            pUI->spinUDMXBs->minimum(1);          
            pUI->spinUDMXBs->maximum(512);
            pUI->spinUDMXBs->value(512);
            pUI->spinUDMXBs->tooltip(_("USBBlock:\nblocksize for transfering data to uDMX.\ndefault 128\nvalid values 1,2,4,8,16...512"));             
          } // Fl_Spinner* o
          { pUI->outSerial = new Fl_Output(40, 255, 140, 20, _("SN"));
            //pUI->outSerial->deactivate();
            pUI->outSerial->hide() ;
            pUI->outSerial->tooltip(_("SN:\nserial number passed with commandline -s <serial>"));             
          } // Fl_Output* o
          { pUI->spinUDMXTb = new Fl_Spinner(275, 215, 65, 20, _("TimeBreak"));
            pUI->spinUDMXTb->minimum(88);
            pUI->spinUDMXTb->maximum(170);
            pUI->spinUDMXTb->value(88);
            pUI->spinUDMXTb->tooltip(_("TimeBreak:\nBreak / Reset time of DMX protocol.\nNeeds uDMX firmware >= 1.3\ndefault 88"));  
            pUI->spinUDMXTb->callback((Fl_Callback*)TimeBreakChange, (void *)pUI); 
          } // Fl_Spinner* o
          { pUI->spinUDMXTm = new Fl_Spinner(275, 235, 65, 20, _("TimeMark"));
            pUI->spinUDMXTm->minimum(8);
            pUI->spinUDMXTm->maximum(170);
            pUI->spinUDMXTm->value(8);
            pUI->spinUDMXTm->tooltip(_("TimeMark:\nfirst mark time of DMX protocol.\nNeeds uDMX firmware >= 1.3\ndefault 8"));             
            pUI->spinUDMXTm->callback((Fl_Callback*)TimeMarkChange, (void *)pUI);             
          } // Fl_Spinner* o
          { pUI->spinUDMXTg = new Fl_Spinner(275, 255, 65, 20, _("TimeIBGap"));
            pUI->spinUDMXTg->minimum(0);
            pUI->spinUDMXTg->maximum(170);
            pUI->spinUDMXTg->value(0);
            pUI->spinUDMXTg->tooltip(_("TimeIBGap:\nGaps of DMX protocol.\nNeeds uDMX firmware >= 1.4\ndefault 0"));             
            pUI->spinUDMXTg->callback((Fl_Callback*)TimeGapChange, (void *)pUI);             
          } // Fl_Spinner* o          
          o->end();
        } // Fl_Group* o
        { Fl_Button * o = new Fl_Button(165, 280, 75, 25, _("Save"));
          o->callback((Fl_Callback*)SettingsSave, (void *)pUI); 
          o->tooltip(_("Save:\nWill save data above. Caution program will freeze for some seconds!"));                      
        } // Fl_Button* o
        o->end();
      } // Fl_Group* o
      { Fl_Group* o = new Fl_Group(0, 26, 425, 289, _("&Info"));
        o->hide();
        { Fl_Help_View * o = new Fl_Help_View(10, 35, 405, 275);
          //o->box(FL_SHADOW_FRAME); // no update when scrolling    
          o->color(FL_BACKGROUND_COLOR) ;          
          o->value(
              "<HTML>\n"
              "<HEAD>\n"
              "<TITLE>uDMX capture - Help</TITLE>\n"
              "</HEAD>\n"
              "<BODY>\n"
              "<H2>uDMX capture</H2>\n" 
              "<H6>This program receives artnet-data and transfers them to an DMX transmitter "
              "based on <a href='http://www.anyma.ch/research/udmx/'>uDMX</a></h6>"
              "<H6>see <a href='http://www.ilLUTZminator.de'>http://www.ilLUTZminator.de</a> "
              "for much more infos</H6><BR><BR>"
              "<H2>License</H2>\n"
              "<H6><a href='http://www.gnu.org/copyleft/gpl.html'>GPLv3</A></H6><BR><BR>"
              "<H2>Made with / thanx to</H2>\n"
              "<H6><a href='http://www.tcpdump.org'>libpcap</a> oder <a href='http://www.winpcap.org/'>winpcap</a><BR>"
              "<a href='http://www.fltk.org/'>http://www.fltk.org/</a><BR>"
              "<a href='http://libusb.sourceforge.net/'>http://libusb.sourceforge.net/</a> or "
              "<a href='http://libusb-win32.sourceforge.net/'>http://libusb-win32.sourceforge.net/</a><BR>"
              "</H6>"              
              "</BODY>\n") ;
        } // Fl_Help_View* o
        o->end();
      } // Fl_Group* o      
      o->end();
    } // Fl_Tabs* o
    o->end();
  } // Fl_Double_Window* o
  w->resizable(w) ;
  return w;
}

// ParseArgs() - get arguments from commandline
// serial number can only be set by commandline so far
void ParseArgs(pUIParams pUI, int argc, char *argv[])
{
	// parse args
	int optc ;

  pUI->bSerial		= false ;
  pUI->cSerial[0] = 0 ;
  pUI->cConfig[0] = 0 ;  

	// parse options 
	while ((optc = getopt (argc, argv, "s:c:")) != EOF)
  {
    // Need looking for serial-Number and
    // then saving to extra file
		switch (optc)
    {
    case 's':
      snprintf(pUI->cSerial, sizeof(pUI->cSerial) - 1, "%s", optarg) ;
      if (strlen(pUI->cSerial))
      {
				char cTmp[256] ;
        pUI->bSerial	= true ;
        
        snprintf(cTmp, sizeof(cTmp), "uDMX capture [%s]", pUI->cSerial);  
        pUI->winMain->copy_label(cTmp) ;

        pUI->outSerial->value(pUI->cSerial) ;
        pUI->outSerial->show() ;
      }
      break;
      
    case 'c':
      snprintf(pUI->cConfig, sizeof(pUI->cConfig) - 1, "%s", optarg) ;
      break;      
    }
	}
  return ;
}

// main() - main program starting program loop
int main(int argc, char **argv)
{
  UserInterface  ui ;
  UIParams       uiParams ;
#ifndef WIN32
  int res ;
#endif  
#ifndef WIN32
  pthread_t           pThread ;
#else
  HANDLE              pThread ;
#endif	

#ifdef FL_LOCK
	Fl::lock() ;
#endif

  // create window
  uiParams.cUI      = &ui ;
  uiParams.winMain  = ui.make_window(&uiParams ) ;

  // make windows icon
#ifndef WIN32
  extern Display *fl_display;

  fl_open_display(); // needed if display has not been previously opened
  Pixmap p, mask;
  XpmCreatePixmapFromData(fl_display, 
                          DefaultRootWindow(fl_display),
                          (char **)uDMX_xpm, &p, &mask, NULL);
  uiParams.winMain->icon((char *)p);
#else
#define USE_RC_ICON
#ifdef USE_RC_ICON
  extern HINSTANCE fl_display;
  
  uiParams.winMain->icon((char *)LoadIcon(fl_display, MAKEINTRESOURCE(IDI_ICON)));
#else    
  HANDLE hImage ;
  
  hImage = LoadImage(NULL, "uDMX.ico", IMAGE_ICON, 32, 32, LR_LOADFROMFILE) ;
  if (hImage)
    uiParams.winMain->icon(hImage);  
#endif
#endif

  // Arguments, Settings, Layout ...
  ParseArgs(&uiParams, argc, argv) ;
  
  ui.SettingsRead(&uiParams) ;

  // dont transfer parameters to mainwindow
  // but we need to use show() with arguments to use the icon
  // thus we reduce the amount of arguments to 1
  //uiParams.winMain->show(argc, argv) ;
  uiParams.winMain->show(1, argv) ;

  // Read layout after show() - otherwise color is not restored
  ui.LayoutRead(&uiParams) ;
  uiParams.winMain->callback(ShutdownQuery, &uiParams) ;

  // show some infos about the program
  uiParams.cUI->Trace(&uiParams, 0, "uDMX capture") ;
  uiParams.cUI->Trace(&uiParams, 0, ".written by ilLU[TZ]minator for") ;
  uiParams.cUI->Trace(&uiParams, 0, "..uDMX hardware developed by http://www.anyma.ch/research/udmx/") ;
  uiParams.cUI->Trace(&uiParams, 0, ".based on") ;
  uiParams.cUI->Trace(&uiParams, 0, "..libpcap / winpcap") ;  
  uiParams.cUI->Trace(&uiParams, 0, "..libusb / libusb-win32 and http://www.fltk.org/") ;  
  uiParams.cUI->Trace(&uiParams, 0, "uDMXCapture started. Use Setup->Verbose to see more or less infos") ;    

  // (Re)Start capturing
  ArtNetRestart(&uiParams) ;

  SLEEP(200) ;

  // create thread for sending the DMX data
#ifndef WIN32
  if ((res = pthread_create(&pThread, NULL, uDMXthread, &uiParams)) != 0)
    uiParams.cUI->Trace(&uiParams, 0, "pthread_create failed. Error %d", res) ;
#else
  pThread = (HANDLE) _beginthread(uDMXthread, 0, (void *)&uiParams) ;
	if ((int)pThread == -1)
    uiParams.cUI->Trace(&uiParams, 0, "_beginthread failed. Error %d", errno) ;	
#endif	

  // set idle function for boring moments
  Fl::add_idle(idle, &uiParams) ;

  // finally begin message loop
	return Fl::run();
}
